/*

  DBOX Module
  Copyright (C) 1994-2000 Daniel Kroening <kroening@handshake.de>

  Purpose: filter management

*/

#include <stdio.h>
#include <string.h>
#include <fstream.h>
#include <stringtools.h>
#include <algorithm>
#include <textprofile.h>

#include "../online/checkuser.h"
#include "challenge.h"
#include "ostatus.h"
#include "webtools.h"
#include "htmltools.h"
#include "filter.h"

string filter_filename()
 {
  return string(status.userprofile.getpath())+"mfilter";
 }

dword get_maxfilter()
 {
  char tempstr[TEXTLEN];

  getuser(&status.userprofile, &status.gruppe, "maxfilter", tempstr);

  return atol(tempstr);
 }

dword count_filter()
 {
  FILE *infile=fopen(filter_filename().c_str(), "rt");
  if(infile==NULL) return 0;

  dword count=0;
  char tempstr[TEXTLEN*2], *tptr;

  while(fgets(tempstr, TEXTLEN-1, infile)!=NULL)
   {
    tptr=strchr(tempstr, '\r'); if(tptr!=NULL) *tptr=0;
    tptr=strchr(tempstr, '\n'); if(tptr!=NULL) *tptr=0;
    if(tempstr[0]!=0) count++;
   }

  fclose(infile);

  return count;
 }

// true: addr is in table
bool check_filter(const char *addr)
 {
  FILE *infile=fopen(filter_filename().c_str(), "rt");
  if(infile==NULL) return 0;

  char t_addr[TEXTLEN], t_alias[TEXTLEN];
  while(readconfigline(infile, t_addr, t_alias, TEXTLEN-1)==0)
    if(stricmp(addr, t_addr)==0)
     {
      fclose(infile);
      return TRUE;
     }

  fclose(infile);

  return FALSE;
 }

struct folder_entryt
 {
  string name;
  int nr;
 };

static bool folder_entry_sort_pr
  (const folder_entryt &a, const folder_entryt &b)
 {
  return stringcasecmp(a.name, b.name)<0;
 }

void dofilter_show(httpt &http, ostream &out)
 {
  bool bg=FALSE;

  out << "<center>\n"
         "<table cellpadding=2 cellspacing=0>\n"
         "<tr bgcolor=\"" << config.webonline_thbgcolor << "\">"
         "<th><font color=\"" << config.webonline_thfcolor 
      << "\">Filterkriterium</font></th><th><font color=\""
      << config.webonline_thfcolor << "\">Vergleichsausdruck</font></th>"
         "<th><font color=\"" << config.webonline_thfcolor 
      << "\">Ziel</font></th>"
      << "<th><font color=\"" << config.webonline_thfcolor
      << "\">Befehle</font></th>\n";

  out << "</tr>\n";

  string challenge;
  get_challenge(challenge, status.userprofile);

  char expr[TEXTLEN], kriterium[TEXTLEN];

  textprofilet profile;

  ifstream in(filter_filename().c_str());  

  if(in)
   {
    while(in && !in.eof())
      if(!profile.readline(in))
       {
        kriterium[0]=0;
        expr[0]=0;

        for(unsigned i=0; i<profile.size(); i++)
         {
          if(strncasecmp(profile.data[i].id, "header.", 7)==0)
           {
            kriterium = profile.data[i].id;
            expr = profile.data[i].data;
           }
          else if(stringcasecmp(profile.data[i].id, "envelope.to")==0)
           {
            kriterium = profile.data[i].id;
            expr = profile.data[i].data;
           }
          else if(stringcasecmp(profile.data[i].id, "envelope.from")==0)
           {
            kriterium = profile.data[i].id;
            expr = profile.data[i].data;
           }
         }

        string box=profile.get("mailbox");
         
        if(kriterium[0]!=0)
         {
    	  out << "<tr bgcolor=\""
              << (bg?config.webonline_tablebgcolor1:config.webonline_tablebgcolor2)
              << "\"><td>" << htmlize_string(expr) << "</td><td>"
              << htmlize_string(kriterium) 	
              << "</td><td>" << htmlize_string(box) << "</td>";

          out << "<td>";

          out << "<a href=\"/filter?command=edit&expr="
              << urlize_string(expr) << "&kriterium="
              << urlize_string(kriterium) 
  	      << "&challenge=" << challenge
              << "\">&auml;ndern</a>";

    	  out << ", <a href=\"/filter?command=del"
                 "&expr=" << urlize_string(expr)
              << "&kriterium=" << urlize_string(kriterium) 
              << "&challenge=" << challenge 
              << "\">l&ouml;schen</a>";

    	  out << "</td>\n";

          bg=!bg;
         }
      }
   }

  out << "<tr bgcolor=\""
      << (bg?config.webonline_tablebgcolor1:config.webonline_tablebgcolor2)
      << "\"><td></td><td></td><td></td>";

  out << "<td><a href=\"/filter?command=new\">neuen Eintrag anlegen</a></td>";

  out << "</tr>\n";

  out << "</table>\n"
         "</center>\n";
 }

void dofilter_select(ostream &out, const string &value)
 {
  out << "<SELECT NAME=\"box\">\n";

  string filename=status.userprofile.getpath();
  filename+="folders";
  
  vector<folder_entryt> folders;
  
  ifstream in(filename.c_str());
    
   {
    folder_entryt folder_entry;
    folder_entry.nr=0;

    if(!in)
     {
      folder_entry.name="INBOX";
      folders.push_back(folder_entry);
     }
    else
     {
      char line[TEXTLEN], *tptr, ch;

      while(in.get(line, TEXTLEN-1))
       {
        in.get(ch); // read \n
        tptr=strchr(line, '\n'); if(tptr!=NULL) *tptr=0;
        tptr=strchr(line, '\r'); if(tptr!=NULL) *tptr=0;

        folder_entry.name=line;
	folders.push_back(folder_entry);
	folder_entry.nr++;
       }
     }
   }

  if(!folders.empty())
    sort(folders.begin()+1, folders.end(), folder_entry_sort_pr);

  for(vector<folder_entryt>::const_iterator it=folders.begin();
      it!=folders.end();
      it++)
    if(it->name!="")
     {
       if(it->name == value)
         out << "<OPTION selected>";
       else
         out << "<OPTION>";

       out << urlize_string(it->name) << "</OPTION>\n";
     }

  out << "</SELECT>";
 }

bool dofilter_check_box(const char *box)
 {
  bool isfolder;
  string filename=status.userprofile.getpath();
  filename+="folders";

  isfolder=TRUE; 
  ifstream in(filename.c_str());
   {
    if(!in)
     {
      if (stricmp("INBOX",box)==0)
         isfolder=FALSE;
     }
    else
     {
      char line[TEXTLEN], *tptr, ch;
      while(in.get(line, TEXTLEN-1))
       {
        in.get(ch); // read \n
        tptr=strchr(line, '\n'); if(tptr!=NULL) *tptr=0;
        tptr=strchr(line, '\r'); if(tptr!=NULL) *tptr=0;
        if (stricmp(line,box)==0)
           isfolder=FALSE; 
       }
     }
   }

  // TRUE liefern, wenn box nicht vorhanden

  return isfolder;
 }

void dofilter_new(httpt &http, ostream &out, const char *expr)
 {
  char filename[DIRLEN], tempstr[DIRLEN], *tmpptr;
  const char *box=http.get_parameter("box");
  const char *kriterium=http.get_parameter("kriterium");

  if(count_filter()>=get_maxfilter())
   {
    startpage(out, "Eintrag hinzuf&uuml;gen");

    out << "<p>Fehler: Die maximale Anzahl an Eintr&auml;gen ist "
           "bereits erreicht.</p>\n"
           "<p>Sie mssen einen Eintrag l&ouml;schen, bevor "
           "ein neuer angelegt werden kann!</p>\n";
   }
  else if(expr!=NULL && box!=NULL && kriterium!=NULL)
   {
    if(verify_challenge(http.getparameter("challenge"), 
       status.userprofile))
     {
      startpage(out, "Eintrag hinzuf&uuml;gen");
      out << "<p>Error: security protocol violated!</p>\n";
      return;
     }

    const char *tptr=expr;
    while(*tptr!=0)
      if(!isalnum(*tptr) && *tptr!='@' && *tptr!='.' &&
         *tptr!='_' && *tptr!='-' && *tptr!='*' && *tptr!='?' )
        break;
      else
        tptr++;

    if(*tptr!=0 || expr[0]==0 )
     {
      startpage(out, "Eintrag hinzuf&uuml;gen");
      out << "<p>Fehler: Filter-Ausdruck ung&uuml;ltig!\n";
      return;
     }

    if(dofilter_check_box(box))
     {
      startpage(out, "Eintrag hinzuf&uuml;gen");
      out << "<p>Fehler: \"" << htmlize_string(box) 
          << "\" ung&uuml;ltig!</p>\n";
      return;
     }

    if(check_filter(expr))
     {
      startpage(out, "Eintrag hinzuf&uuml;gen");
      out << "<p>Fehler: Ausdruck ist bereits gefiltert!</p>\n";
      return;
     }

    FILE *infile=fopen(filter_filename().c_str(), "rt");
    snprintf(filename, DIRLEN, "%s.tmp", filter_filename().c_str());
    FILE *outfile=fopen(filename, "wt");
    if(outfile==NULL)
     {
       startpage(out, "Eintrag hinzuf&uuml;gen");
       out << "<p>Error: failed to open tmp filterfile!</p>\n";
       return;
     } 

    if(infile!=NULL)
     {
      while(fgets(tempstr, DIRLEN-1, infile)!=NULL)
       {
        tmpptr=strchr(tempstr, '\r'); if(tmpptr!=NULL) *tmpptr=0;
        tmpptr=strchr(tempstr, '\n'); if(tmpptr!=NULL) *tmpptr=0;
        if(tempstr[0]!=0 && stricmp(tempstr, "mailbox:INBOX")!=0)
        fprintf(outfile, "%s\n", tempstr);
       }

      fclose(infile);
     }

    fprintf(outfile, "break=1;%s:%s;mailbox:%s\n", kriterium, expr, box);
    fprintf(outfile, "mailbox:INBOX\n");
    fclose(outfile);
    rename(filename, filter_filename().c_str());
    
    redirect(out, http.url);
   }
  else
   {
    startpage(out, "Eintrag hinzuf&uuml;gen");

    out <<
      "<FORM ACTION=\"/filter\">\n"
      "<center>\n"
      "<table>\n"
      "<tr><th colspan=2>"
      "<font size=+1><u>Eintrag anlegen</u></font></th></tr>\n"
      "<tr><th align=right>Kriterium:</th>\n"
      "<SELECT name=kriterium>\n"
      "<OPTION value=\"envelope.to\">To (Envelope)</OPTION>\n"
      "<OPTION value=\"envelope.from\">From (Envelope)</OPTION>\n"
      "<OPTION value=\"header.EMP\">Empf&auml;nger</OPTION>\n"
      "<OPTION value=\"header.ABS\">Absender</OPTION>\n"
      "<OPTION value=\"header.BET\">Betreff</OPTION>\n"
      "<OPTION value=\"header.LEN\">L&auml;nge</OPTION>\n"
      "<OPTION value=\"header.ROT\">Zustell-Pfad</OPTION>\n"
      "<OPTION value=\"header.MID\">Message-ID</OPTION>\n"
      "<OPTION value=\"header.ORG\">Organisation</OPTION>\n"
      "<OPTION value=\"header.ANTWORT-AN\">Antwort-an</OPTION>\n"
      "<OPTION value=\"header.ZUSAMMENFASSUNG\">Zusammenfassung</OPTION>\n"
      "<OPTION value=\"header.MAILER\">Mailer</OPTION>\n" 
      "</SELECT>\n"
      "<tr><th align=right>Filterausdruck:</th>\n"
      "<td><input name=\"expr\" size=\"40\"></td></tr>\n"
      "<tr><th align=right>Ziel-Box:</th>\n"
      "<td>";

    dofilter_select(out, "");

    string challenge;
    get_challenge(challenge, status.userprofile);

    out << 
      "</td></tr>\n"
      "<tr><td colspan=2><hr size=1 noshade></td></tr>\n"
      "<tr><td></td><td>\n"
      "<INPUT TYPE=submit value=\"Eintrag anlegen\"></td></tr>\n"
      "<INPUT TYPE=hidden name=\"command\" value=\"new\">\n"
      "<INPUT TYPE=hidden name=\"challenge\" value=\"" << challenge << "\">\n"
      "</table>\n"
      "</center\n"
      "</form>\n"
      "<p>\n";
   }
 }

void dofilter_del(httpt &http, ostream &out, const char *expr)
 {
  const char *kriterium=http.getparameter("kriterium");
  if(expr!=NULL)
   {
    if(verify_challenge(http.getparameter("challenge"), 
       status.userprofile))
     {
      startpage(out, "Eintrag l&ouml;schen");
      out << "<p>Error: security protocol violated!</p>\n";
      return;
     }

    string filename=filter_filename();
    string tmp_filename=filename+".tmp";
    FILE *outfile=fopen(tmp_filename.c_str(), "wt");
    if(outfile==NULL)  return; 

    char t_expr[TEXTLEN], t_kriterium[TEXTLEN];
    textprofilet profile;

    ifstream in(filter_filename().c_str());
    if(in)
     {
      while(in && !in.eof())
        if(!profile.readline(in))
         {
          t_kriterium[0]=0;
          t_expr[0]=0;
          for (unsigned i=0; i<profile.size(); i++)
           {
            if(strncasecmp(profile.data[i].id, "header.", 7)==0)
             {
               t_kriterium = profile.data[i].id;
               t_expr = profile.data[i].data;
             }
            else if(stringcasecmp(profile.data[i].id, "envelope.to")==0)
             {
               t_kriterium = profile.data[i].id;
               t_expr = profile.data[i].data;
             }
            else if(stringcasecmp(profile.data[i].id, "envelope.from")==0)
             {
               t_kriterium = profile.data[i].id;
               t_expr = profile.data[i].data;
             }
           }
  
          const char *t_box=profile.get("mailbox");

          if((stricmp(expr, t_expr)!=0)||(stricmp(kriterium, t_kriterium)!=0))
           {
            if(t_kriterium[0]!=0)
              fprintf(outfile, "break=1;%s:%s;mailbox=%s\n", t_kriterium, t_expr, t_box);
            else
              fprintf(outfile, "mailbox=INBOX\n");
           }
         }

        fclose(outfile);
        rename(tmp_filename.c_str(), filename.c_str());
     }
   }

  redirect(out, http.url);
 }

void dofilter_edit(httpt &http, ostream &out, const char *expr)
 {
  const char *kriterium=http.get_parameter("kriterium");
  const char *box=http.get_parameter("box");
  const char *oldexpr=http.get_parameter("oldexpr");

  if(expr==NULL)
   {
    startpage(out, "ERROR 1");
    return;
   }

  if(kriterium==NULL)
   {
    startpage(out, "ERROR 2");
    return;
   }
 
  if(box!=NULL)
   {
    if(verify_challenge(http.getparameter("challenge"), 
       status.userprofile))
     {
      startpage(out, "Eintrag &auml;ndern");
      out << "<p>Error: security protocol violated!</p>\n";
      return;
     }

    if(dofilter_check_box(box))
     {
      startpage(out, "Eintrag &auml;ndern");
      out << "<p>Fehler: \"" << htmlize_string(box) 
          << "\" ung&uuml;ltig!</p>\n";
      return;
     }

    string filename=filter_filename();
    string tmp_filename=filename+".tmp"; // todo
    FILE *outfile=fopen(tmp_filename.c_str(), "wt");
    if(outfile==NULL) return;  // todo

    char t_expr[TEXTLEN], t_kriterium[TEXTLEN];
    textprofilet profile;
    ifstream in(filter_filename().c_str());
    if(in)
     {
      while(in && !in.eof())
        if(!profile.readline(in))
         {
          t_kriterium[0]=0;
          t_expr[0]=0;
          for (unsigned i=0; i<profile.size(); i++)
           {
            if(strncasecmp(profile.data[i].id, "header.", 7)==0)
             {
              t_kriterium = profile.data[i].id;
              t_expr = profile.data[i].data;
             }
            else if(stringcasecmp(profile.data[i].id, "envelope.to")==0)
             {
              t_kriterium = profile.data[i].id;
              t_expr = profile.data[i].data;
             }
            else if(stringcasecmp(profile.data[i].id, "envelope.from")==0)
             {
              t_kriterium = profile.data[i].id;
              t_expr = profile.data[i].data;
             }
           }

          const char *t_box=profile.get("mailbox");   

          if((stricmp(oldexpr, t_expr)!=0)||(stricmp(kriterium, t_kriterium)!=0))
           { 
            if(t_kriterium[0]!=0)
              fprintf(outfile, "break=1;%s:%s;mailbox=%s\n", t_kriterium, t_expr, t_box); 
            else 
              fprintf(outfile, "mailbox=INBOX\n");
           }
          else
            fprintf(outfile, "break=1;%s:%s;mailbox=%s\n", t_kriterium, expr, box);
         }

      fclose(outfile);
      rename(tmp_filename.c_str(), filename.c_str());
     }

    redirect(out, http.url);
   }
  else
   {
    startpage(out, "Eintrag &auml;ndern (Formular)");

    out <<
      "<FORM ACTION=\"/filter\">\n"
      "<center>\n"
      "<table>\n"
      "<tr><th colspan=2><font size=+1><u>Eintrag &auml;ndern\n"
      "</u></font></th></tr>\n"
      "<tr><th align=right>Filterkriterium:</th>\n"
      "<td><code>" << htmlize_string(kriterium) << "</code>"
       << "</td></tr>\n"
      "<tr><th align=right>Vergleichsausdruck:</th>\n"
      "<td><input name=\"expr\" size=\"40\" value=\""
       << htmlize_string(expr) << "\"></td></tr>\n"
      "<tr><th align=right>Ziel-Box:</th><td>\n";

    dofilter_select(out, "");

    string challenge;
    get_challenge(challenge, status.userprofile);

    out << 
      "</td></tr>\n"
      "<tr><td colspan=2><hr size=1 noshade></td></tr>\n"
      "<tr><td></td><td>"
      "<INPUT TYPE=submit value=\"Eintrag &auml;ndern\"></td></tr>\n"
      "<INPUT TYPE=hidden name=\"command\" value=\"edit\">\n"
      "<INPUT TYPE=hidden name=\"kriterium\" value=\""
      << htmlize_string(kriterium) << "\">\n"
      "<INPUT TYPE=hidden name=\"oldexpr\" value=\""
      << htmlize_string(expr) << "\">\n"
      "<INPUT TYPE=hidden name=\"challenge\" value=\""
      << challenge << "\">\n"
      "</table>\n"
      "</center\n"
      "</form>\n"
      "<p>\n";
   }

  endpage(out);
 }

void dofilter(httpt &http, ostream &out)
 {
  const char *command, *expr;

  if((status.userprofile|"alias")[0]!=0)
   {
    startpage(out, "Filter Tabelle");
    out << "Zugriff verweigert!\n";
   }
  else
   {
    command=http.getparameter("command");
    expr=http.getparameter("expr");

    if(command==NULL)
     {
      startpage(out, "Filter Tabelle");
      dofilter_show(http, out);
     }
    else if(stricmp(command, "new")==0)
      dofilter_new(http, out, expr);
    else if(stricmp(command, "del")==0)
      dofilter_del(http, out, expr);
    else if(stricmp(command, "edit")==0)
      dofilter_edit(http, out, expr);
    else
     {
      startpage(out, "Error");
      out << "Unknown command.\n";
     }
   }

  showcopyright(out);
  endpage(out);
 }
